@@ -62,7 +62,14 @@
 #define MATE_TOUCHPAD_SCHEMA             "org.mate.peripherals-touchpad"
 #define KEY_TOUCHPAD_DISABLE_W_TYPING    "disable-while-typing"
 #ifdef HAVE_X11_EXTENSIONS_XINPUT_H
+ 
+#define KEY_TOUCHPAD_TWO_FINGER_CLICK    "two-finger-click"
+#define KEY_TOUCHPAD_THREE_FINGER_CLICK  "three-finger-click"
+#define KEY_TOUCHPAD_NATURAL_SCROLL      "natural-scroll"
 #define KEY_TOUCHPAD_TAP_TO_CLICK        "tap-to-click"
+#define KEY_TOUCHPAD_ONE_FINGER_TAP      "tap-button-one-finger"
+#define KEY_TOUCHPAD_TWO_FINGER_TAP      "tap-button-two-finger"
+#define KEY_TOUCHPAD_THREE_FINGER_TAP    "tap-button-three-finger"
 #define KEY_TOUCHPAD_SCROLL_METHOD       "scroll-method"
 #define KEY_TOUCHPAD_PAD_HORIZ_SCROLL    "horiz-scroll-enabled"
 #define KEY_TOUCHPAD_ENABLED             "touchpad-enabled"
@@ -92,7 +99,9 @@ static void     msd_mouse_manager_init        (MsdMouseManager      *mouse_manag
 static void     msd_mouse_manager_finalize    (GObject             *object);
 static void     set_mouse_settings            (MsdMouseManager      *manager);
 #ifdef HAVE_X11_EXTENSIONS_XINPUT_H
-static int      set_tap_to_click              (gboolean state, gboolean left_handed);
+static void     set_tap_to_click              (MsdMouseManager * manager);
+static void     set_click_actions             (MsdMouseManager * manager);
+static void     set_natural_scroll            (MsdMouseManager * manager); 
 #endif
 
 G_DEFINE_TYPE (MsdMouseManager, msd_mouse_manager, G_TYPE_OBJECT)
@@ -300,7 +309,7 @@ touchpad_has_single_button (XDevice *device)
 
 
 static void
-set_xinput_devices_left_handed (gboolean left_handed)
+set_xinput_devices_left_handed (MsdMouseManager * manager, gboolean left_handed)
 {
         XDeviceInfo *device_info;
         gint n_devices;
@@ -334,7 +343,7 @@ set_xinput_devices_left_handed (gboolean left_handed)
                         gboolean single_button = touchpad_has_single_button (device);
 
                         if (tap && !single_button)
-                                set_tap_to_click (tap, left_handed);
+                                set_tap_to_click (manager);
                         XCloseDevice (GDK_DISPLAY_XDISPLAY(gdk_display_get_default()), device);
                         g_object_unref (settings);
 
@@ -432,7 +441,7 @@ set_left_handed (MsdMouseManager *manager,
                 /* When XInput support is available, never set the
                  * button ordering on the core pointer as that would
                  * revert the changes we make on the devices themselves */
-                set_xinput_devices_left_handed (left_handed);
+                set_xinput_devices_left_handed (manager, left_handed);
                 return;
         }
 #endif
@@ -570,13 +579,25 @@ set_middle_button (MsdMouseManager *manager,
                 XFreeDeviceList (device_info);
 }
 
+static gboolean
+have_program_in_path (const char *name)
+{
+        gchar *path;
+        gboolean result;
+
+        path = g_find_program_in_path (name);
+        result = (path != NULL);
+        g_free (path);
+        return result;
+}
+
 static int
 set_disable_w_typing (MsdMouseManager *manager, gboolean state)
 {
 
         if (state && touchpad_is_present ()) {
                 GError *error = NULL;
-                char *args[5];
+                char *args[6];
 
                 if (manager->priv->syndaemon_spawned)
                         return 0;
@@ -584,10 +605,11 @@ set_disable_w_typing (MsdMouseManager *manager, gboolean state)
                 args[0] = "syndaemon";
                 args[1] = "-i";
                 args[2] = "0.5";
-                args[3] = "-k";
-                args[4] = NULL;
+                args[3] = "-K";
+                args[4] = "-R";
+                args[5] = NULL;
 
-                if (!g_find_program_in_path (args[0]))
+                if (!have_program_in_path (args[0]))
                         return 0;
 
                 g_spawn_async (g_get_home_dir (), args, NULL,
@@ -612,8 +634,8 @@ set_disable_w_typing (MsdMouseManager *manager, gboolean state)
 }
 
 #ifdef HAVE_X11_EXTENSIONS_XINPUT_H
-static int
-set_tap_to_click (gboolean state, gboolean left_handed)
+static void
+set_tap_to_click (MsdMouseManager * manager)
 {
         int numdevices, i, format, rc;
         unsigned long nitems, bytes_after;
@@ -623,12 +645,18 @@ set_tap_to_click (gboolean state, gboolean left_handed)
         Atom prop, type;
 
         if (devicelist == NULL)
-                return 0;
+                return;
 
         prop = XInternAtom (GDK_DISPLAY_XDISPLAY(gdk_display_get_default()), "Synaptics Tap Action", False);
 
         if (!prop)
-                return 0;
+                return;
+
+        gboolean state = g_settings_get_boolean (manager->priv->settings_touchpad, KEY_TOUCHPAD_TAP_TO_CLICK);
+        gboolean left_handed = g_settings_get_boolean (manager->priv->settings_mouse, KEY_MOUSE_LEFT_HANDED);
+        gint one_finger_tap = g_settings_get_int (manager->priv->settings_touchpad, KEY_TOUCHPAD_ONE_FINGER_TAP);
+        gint two_finger_tap = g_settings_get_int (manager->priv->settings_touchpad, KEY_TOUCHPAD_TWO_FINGER_TAP);
+        gint three_finger_tap = g_settings_get_int (manager->priv->settings_touchpad, KEY_TOUCHPAD_THREE_FINGER_TAP);
 
         for (i = 0; i < numdevices; i++) {
                 if ((device = device_is_touchpad (devicelist[i]))) {
@@ -637,12 +665,19 @@ set_tap_to_click (gboolean state, gboolean left_handed)
                                                 False, XA_INTEGER, &type, &format, &nitems,
                                                 &bytes_after, &data);
 
+                        if (one_finger_tap > 3 || one_finger_tap < 1)
+                                one_finger_tap = 1;
+                        if (two_finger_tap > 3 || two_finger_tap < 1)
+                                two_finger_tap = 3;
+                        if (three_finger_tap > 3 || three_finger_tap < 1)
+                                three_finger_tap = 2;
+                        
                         if (rc == Success && type == XA_INTEGER && format == 8 && nitems >= 7)
                         {
                                 /* Set RLM mapping for 1/2/3 fingers*/
-                                data[4] = (state) ? ((left_handed) ? 3 : 1) : 0;
-                                data[5] = (state) ? ((left_handed) ? 1 : 3) : 0;
-                                data[6] = (state) ? 2 : 0;
+                                data[4] = (state) ? ((left_handed) ? (4-one_finger_tap) : one_finger_tap) : 0;
+                                data[5] = (state) ? ((left_handed) ? (4-two_finger_tap) : two_finger_tap) : 0;
+                                data[6] = (state) ? three_finger_tap : 0;
                                 XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY(gdk_display_get_default()), device, prop, XA_INTEGER, 8,
                                                         PropModeReplace, data, nitems);
                         }
@@ -658,7 +693,106 @@ set_tap_to_click (gboolean state, gboolean left_handed)
         }
 
         XFreeDeviceList (devicelist);
-        return 0;
+}
+
+static void
+set_click_actions (MsdMouseManager * manager)
+{
+        int numdevices, i, format, rc;
+        unsigned long nitems, bytes_after;
+        XDeviceInfo *devicelist = XListInputDevices (GDK_DISPLAY_XDISPLAY(gdk_display_get_default()), &numdevices);
+        XDevice * device;
+        unsigned char* data;
+        Atom prop, type;
+
+        if (devicelist == NULL)
+                return;
+
+        prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Click Action", False);
+        if (!prop)
+                return;
+
+        gint enable_two_finger_click = g_settings_get_int (manager->priv->settings_touchpad, KEY_TOUCHPAD_TWO_FINGER_CLICK);
+        gint enable_three_finger_click = g_settings_get_int (manager->priv->settings_touchpad, KEY_TOUCHPAD_THREE_FINGER_CLICK);
+
+        for (i = 0; i < numdevices; i++) {
+                if ((device = device_is_touchpad (devicelist[i]))) {
+                        g_debug ("setting click action to click on %s", devicelist[i].name);
+                        gdk_error_trap_push ();
+                        rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), device, prop, 0, 2,
+                                         False, XA_INTEGER, &type, &format, &nitems,
+                                         &bytes_after, &data);
+
+                        if (rc == Success && type == XA_INTEGER && format == 8 && nitems >= 3) {
+                                data[0] = 1;
+                                data[1] = enable_two_finger_click;
+                                data[2] = enable_three_finger_click;
+                                XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), device, prop, XA_INTEGER, 8, PropModeReplace, data, nitems);
+                        }
+
+                        if (rc == Success)
+                                XFree (data);
+                        XCloseDevice (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), device);
+                        if (gdk_error_trap_pop ()) {
+                                g_warning ("Error in setting click actions on \"%s\"", devicelist[i].name);
+                                continue;
+                        }
+                }
+        }
+        XFreeDeviceList (devicelist);
+}
+
+static void
+set_natural_scroll (MsdMouseManager * manager)
+{
+        int numdevices, i, format, rc;
+        unsigned long nitems, bytes_after;
+        XDeviceInfo *devicelist = XListInputDevices (GDK_DISPLAY_XDISPLAY(gdk_display_get_default()), &numdevices);
+        XDevice * device;
+        unsigned char* data;
+        glong *ptr;
+        Atom prop, type;
+
+        if (devicelist == NULL)
+                return;
+
+        prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Scrolling Distance", False);
+        if (!prop)
+                return;
+
+        gboolean natural_scroll = g_settings_get_boolean (manager->priv->settings_touchpad, KEY_TOUCHPAD_NATURAL_SCROLL);
+
+        for (i = 0; i < numdevices; i++) {
+                if ((device = device_is_touchpad (devicelist[i]))) {
+                        g_debug ("Trying to set %s for \"%s\"", natural_scroll ? "natural (reverse) scroll" : "normal scroll", devicelist[i].name);
+                        gdk_error_trap_push ();
+                        rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), device, prop, 0, 2,
+                                         False, XA_INTEGER, &type, &format, &nitems,
+                                         &bytes_after, &data);
+
+                        if (rc == Success && type == XA_INTEGER && format == 32 && nitems >= 2) {
+                                ptr = (glong *) data;
+                                if (natural_scroll) {
+                                        ptr[0] = -abs(ptr[0]);
+                                        ptr[1] = -abs(ptr[1]);
+                                } else {
+                                        ptr[0] = abs(ptr[0]);
+                                        ptr[1] = abs(ptr[1]);
+                                }
+                                
+                                XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), device, prop, XA_INTEGER, 32, PropModeReplace, data, nitems);
+                        }
+
+                        if (rc == Success)
+                                XFree (data);
+                        XCloseDevice (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), device);
+                        if (gdk_error_trap_pop ()) {
+                                g_warning ("Error in setting natural scroll on \"%s\"", devicelist[i].name);
+                                continue;
+                        }
+                }
+        }
+        XFreeDeviceList (devicelist);
 }
 
 static int
@@ -933,13 +1067,15 @@ set_mouse_settings (MsdMouseManager *manager)
         set_left_handed (manager, left_handed);
         set_motion_acceleration (manager, g_settings_get_double (manager->priv->settings_mouse, KEY_MOUSE_MOTION_ACCELERATION));
         set_motion_threshold (manager, g_settings_get_int (manager->priv->settings_mouse, KEY_MOUSE_MOTION_THRESHOLD));
-	set_middle_button (manager, g_settings_get_boolean (manager->priv->settings_mouse, KEY_MIDDLE_BUTTON_EMULATION));
+        set_middle_button (manager, g_settings_get_boolean (manager->priv->settings_mouse, KEY_MIDDLE_BUTTON_EMULATION));
 
         set_disable_w_typing (manager, g_settings_get_boolean (manager->priv->settings_touchpad, KEY_TOUCHPAD_DISABLE_W_TYPING));
 #ifdef HAVE_X11_EXTENSIONS_XINPUT_H
-        set_tap_to_click (g_settings_get_boolean (manager->priv->settings_touchpad, KEY_TOUCHPAD_TAP_TO_CLICK), left_handed);
+        set_tap_to_click (manager);
+        set_click_actions (manager);
         set_edge_scroll (g_settings_get_int (manager->priv->settings_touchpad, KEY_TOUCHPAD_SCROLL_METHOD));
         set_horiz_scroll (g_settings_get_boolean (manager->priv->settings_touchpad, KEY_TOUCHPAD_PAD_HORIZ_SCROLL));
+        set_natural_scroll (manager);
         set_touchpad_enabled (g_settings_get_boolean (manager->priv->settings_touchpad, KEY_TOUCHPAD_ENABLED));
 #endif
 }
@@ -957,15 +1093,24 @@ mouse_callback (GSettings          *settings,
                 set_motion_threshold (manager, g_settings_get_int (settings, key));
         } else if (g_strcmp0 (key, KEY_TOUCHPAD_DISABLE_W_TYPING) == 0) {
                 set_disable_w_typing (manager, g_settings_get_boolean (settings, key));
-	} else if (g_str_equal (key, KEY_MIDDLE_BUTTON_EMULATION)) {
-	        set_middle_button (manager, g_settings_get_boolean (settings, key));
+	    } else if (g_str_equal (key, KEY_MIDDLE_BUTTON_EMULATION)) {
+	            set_middle_button (manager, g_settings_get_boolean (settings, key));
 #ifdef HAVE_X11_EXTENSIONS_XINPUT_H
         } else if (g_strcmp0 (key, KEY_TOUCHPAD_TAP_TO_CLICK) == 0) {
-                set_tap_to_click (g_settings_get_boolean (settings, key),
-                                  g_settings_get_boolean (manager->priv->settings_mouse, KEY_MOUSE_LEFT_HANDED));
+                set_tap_to_click (manager);
+        } else if (g_str_equal (key, KEY_TOUCHPAD_TWO_FINGER_CLICK) || g_str_equal (key, KEY_TOUCHPAD_THREE_FINGER_CLICK)) {
+                set_click_actions(manager);
+        } else if (g_strcmp0 (key, KEY_TOUCHPAD_ONE_FINGER_TAP) == 0) {
+                set_tap_to_click (manager);
+        } else if (g_strcmp0 (key, KEY_TOUCHPAD_TWO_FINGER_TAP) == 0) {
+                set_tap_to_click (manager);
+        } else if (g_strcmp0 (key, KEY_TOUCHPAD_THREE_FINGER_TAP) == 0) {
+                set_tap_to_click (manager);
         } else if (g_strcmp0 (key, KEY_TOUCHPAD_SCROLL_METHOD) == 0) {
                 set_edge_scroll (g_settings_get_int (settings, key));
                 set_horiz_scroll (g_settings_get_boolean (settings, KEY_TOUCHPAD_PAD_HORIZ_SCROLL));
+        } else if (g_str_equal (key, KEY_TOUCHPAD_NATURAL_SCROLL)) {
+                set_natural_scroll (manager);
         } else if (g_strcmp0 (key, KEY_TOUCHPAD_PAD_HORIZ_SCROLL) == 0) {
                 set_horiz_scroll (g_settings_get_boolean (settings, key));
 #endif
@@ -1023,10 +1168,11 @@ msd_mouse_manager_idle_cb (MsdMouseManager *manager)
 
         set_disable_w_typing (manager, g_settings_get_boolean (manager->priv->settings_touchpad, KEY_TOUCHPAD_DISABLE_W_TYPING));
 #ifdef HAVE_X11_EXTENSIONS_XINPUT_H
-        set_tap_to_click (g_settings_get_boolean (manager->priv->settings_touchpad, KEY_TOUCHPAD_TAP_TO_CLICK),
-                          g_settings_get_boolean (manager->priv->settings_mouse, KEY_MOUSE_LEFT_HANDED));
+        set_tap_to_click (manager);
+        set_click_actions (manager);
         set_edge_scroll (g_settings_get_int (manager->priv->settings_touchpad, KEY_TOUCHPAD_SCROLL_METHOD));
         set_horiz_scroll (g_settings_get_boolean (manager->priv->settings_touchpad, KEY_TOUCHPAD_PAD_HORIZ_SCROLL));
+        set_natural_scroll (manager);
         set_touchpad_enabled (g_settings_get_boolean (manager->priv->settings_touchpad, KEY_TOUCHPAD_ENABLED));
 #endif
 
